package czerwony.kapturek.jbspell;

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.event.ActionEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;

import com.borland.primetime.editor.*;
import com.borland.primetime.ide.*;
import com.borland.primetime.node.*;

/**
 * Action for the spell checker. This class encapsulates some classes that do
 * actual spell checking job and provides a few utility methods.
 */
class Actions implements ChangeListener
{
    private final static String TXT_ACTIONPREFIX = "jbSpell-";
    private final static String TXT_COLOR_GOOD = "Spell checker - OK";
    private final static String TXT_COLOR_WRONG = "Spell checker - Error";

    public static SpellAction ACTION_CheckWord;
    public static SpellAction ACTION_CheckFile;

    private EditorAction[] actions;

    private static Color colorGood = new Color(0, 255, 0);
    private static Color colorWrong = new Color(255, 128, 128);

    /**
     * Creates and stores in the table one instance of all EditorAction based
     * inner classes.
     */
    Actions()
    {
        Vector v = new Vector();

        ACTION_CheckWord = new CheckWordAction();
        v.add(ACTION_CheckWord);

        ACTION_CheckFile = new CheckFileAction();
        v.add(ACTION_CheckFile);

        actions = (EditorAction[])v.toArray(new EditorAction[0]);

        MasterStyleContext styleCtx = EditorManager.getStyleContext();

        Style ok = styleCtx.addStyle(TXT_COLOR_GOOD, null);
        ok.addAttribute(Style.NameAttribute, TXT_COLOR_GOOD);
        StyleConstants.setBackground(ok, colorGood);
        ok.addChangeListener(this);
        colorGood = StyleConstants.getBackground(ok);

        Style wrong = styleCtx.addStyle(TXT_COLOR_WRONG, null);
        wrong.addAttribute(Style.NameAttribute, TXT_COLOR_WRONG);
        StyleConstants.setBackground(wrong, colorWrong);
        wrong.addChangeListener(this);
        colorWrong = StyleConstants.getBackground(wrong);
    }


    /**
     * Called when user changes the colors to be used for highlighting.
     */
    public void stateChanged(ChangeEvent e)
    {
        Style style = (Style)e.getSource();
        if (style.getAttribute(Style.NameAttribute).equals(TXT_COLOR_WRONG))
            colorWrong = StyleConstants.getBackground(style);
        else
            colorGood = StyleConstants.getBackground(style);
    }


    public EditorAction[] getActions()
    {
        return actions;
    }


    /**
     * Utility function for temporarily highlighting text in EditorPane.
     *
     * @param txt   Text control containing the text to be highlighted.
     * @param beg   Position of the beginning of the highlighted text.
     * @param end   Ending position of the highlighted text.
     * @param c     Color of the highlight background.
     *
     * @throws BadLocationException Thrown when one beg or end parameters have
     *              illegal values.
     */
    private void temporarilyHighlightText(EditorPane txt, int beg, int end,
                                          Color c)
        throws BadLocationException
    {
        Highlighter highlighter = txt.getHighlighter();
        Object highlight =
            highlighter.addHighlight(beg, end, new DefaultHighlighter.DefaultHighlightPainter(c));
        new javax.swing.Timer(300, new HighlightRemover(txt, highlight)).start();
    }


    class HighlightRemover implements ActionListener
    {
        private Object hi;
        private EditorPane ed;

        HighlightRemover(EditorPane ed, Object hi)
        {
            this.ed = ed;
            this.hi = hi;
        }


        public void actionPerformed(ActionEvent ev)
        {
            ed.getHighlighter().removeHighlight(hi);
        }
    }


    private void replaceWord(EditorPane editor, int beg, int end, String newWord)
    {
        try
        {
            Document doc = editor.getDocument();
            doc.remove(beg, end - beg);
            doc.insertString(beg, newWord, null);
        }
        catch (BadLocationException ex)
        {
        }
    }


    private String skipDigits(String word)
    {
        int i = 0;
        for (; i < word.length(); i++)
        {
            if (!Character.isDigit(word.charAt(i)))
            {
                break;
            }
        }

        return word.substring(i);
    }


    private boolean isJavaDocTag(EditorPane ed, int beg)
        throws BadLocationException
    {
        return ed.getText(beg - 1, 1).equals("@");
    }


    abstract class SpellAction extends EditorAction
    {
        SpellAction(String name, String descr)
        {
            super(TXT_ACTIONPREFIX + name);

            putValue(LONG_DESCRIPTION, descr);
            putValue("ActionGroup", "Spell Checker (jbSpell)");
        }

        public void actionPerformed(ActionEvent ev)
        {
            try
            {
                EditorPane target = getEditorTarget(ev);
                if (target != null)
                {
                    onAction(target);
                }
            }
            catch (SpellCheckerException ex)
            {
                JOptionPane.showMessageDialog(Browser.getActiveBrowser(), ex.getMessage());
                ex.tryToRecover();
            }
        }

        protected abstract void onAction(EditorPane target) throws SpellCheckerException;
    }


    public class CheckWordAction extends SpellAction
    {
        public CheckWordAction()
        {
            super("check-word", "Check spelling of the word under (or before) cursor.");
        }


        public void onAction(EditorPane target) throws SpellCheckerException
        {
            try
            {
                int p0 = target.getCaretPosition();
                int beg = TextUtilities.getWordStart(target,
                                                     p0);           // , false);
                int end = TextUtilities.getWordEnd(target, beg);    // , false);
                String word = skipDigits(target.getText(beg, end - beg));
                if (word.length() == 0)
                {
                    return;
                }

                AbstractChecker checker = SpellCheckerTool.getChecker();
                if (!checker.isInDictionary(word))
                {
                    Object hi =
                        target.getHighlighter()
                            .addHighlight(beg, end, new DefaultHighlighter
                                .DefaultHighlightPainter(colorWrong));

                    String[] suggestions = checker.getSuggestions(word);
                    if (Settings.useNativeMenus.getBoolean())
                    {
                        displayNativeSuggestionsMenu(word, suggestions,
                                                     target, beg, end, hi);
                    }
                    else
                    {
                        displaySuggestionsMenu(word, suggestions, target,
                                               beg, end, hi);
                    }
                }
                else
                {
                    temporarilyHighlightText(target, beg, end, colorGood);
                }
            }
            catch (BadLocationException ex)
            {
            }
        }


        private void displaySuggestionsMenu(String word, String[] suggestions,
                                            EditorPane ed, int beg, int end,
                                            Object hi)
        {
            try
            {
                Rectangle rect = ed.modelToView(beg);
                rect.width = ed.modelToView(end).x - rect.x;

                JPopupMenu menu = new JPopupMenu();
                if ((suggestions != null) && (suggestions.length > 0))
                {
                    ActionListener al = new ReplaceWordActionListener(ed, beg,
                                            end);
                    for (int i = 0; i < suggestions.length;
                            menu.add(suggestions[i++]).addActionListener(al));
                }
                else
                {
                    JMenuItem item = new JMenuItem("(no spelling suggestions)");
                    item.setEnabled(false);
                    menu.add(item);
                }

                menu.addSeparator();

//                JMenuItem itemIgnore = new JMenuItem("Ignore all");
//                itemIgnore.addActionListener(new IgnoreAllActionListener(word));
//                itemIgnore.setMnemonic('I');
                JMenuItem itemAdd = new JMenuItem("Add to dictionary");
                itemAdd.addActionListener(new AddToDictionaryActionListener(word));
                itemAdd.setMnemonic('A');
//                menu.add(itemIgnore);
                menu.add(itemAdd);
                menu.addPopupMenuListener(new SuggestionsMenuListener(ed, hi));
                ed.add(menu);

                Rectangle menuRect =
                    new Rectangle(0, 0, menu.getPreferredSize().width,
                                  menu.getPreferredSize().height);
                Point p = calculateMenuPosition(menuRect, rect, ed);
                menu.show(ed, p.x, p.y);
            }
            catch (BadLocationException ex)
            {
            }
        }


        private Point calculateMenuPosition(Rectangle rectMenu,
                                            Rectangle rectWord, JComponent comp)
        {
            Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
            Point p = rectWord.getLocation();
            SwingUtilities.convertPointToScreen(p, comp);

            p.y = p.y + rectWord.height;

            if ((screen.height - p.y) < rectMenu.height)
            {
                p.y = p.y - rectMenu.height - rectWord.height;
            }

            if ((screen.width - p.x - rectWord.width) < rectMenu.width)
            {
                p.x = p.x + rectWord.width - rectMenu.width;
            }

            SwingUtilities.convertPointFromScreen(p, comp);

            return p;
        }


        /**
         * Listener for the spell checker menu. Needed only to remove highlight
         * from the incorrectly spelled word when the menu disappears.
         * Unfortunately there is no adapter class for this one.
         */
        class SuggestionsMenuListener implements PopupMenuListener
        {
            private EditorPane editor;
            private Object highlight;

            SuggestionsMenuListener(EditorPane ed, Object hi)
            {
                editor = ed;
                highlight = hi;
            }


            public void popupMenuWillBecomeVisible(PopupMenuEvent ev)
            {
            }


            public void popupMenuWillBecomeInvisible(PopupMenuEvent ev)
            {
                editor.getHighlighter().removeHighlight(highlight);
            }


            public void popupMenuCanceled(PopupMenuEvent ev)
            {
            }
        }


        private void displayNativeSuggestionsMenu(String word,
                                                  String[] suggestions,
                                                  EditorPane ed, int beg,
                                                  int end, Object hi)
        {
            try
            {
                Rectangle rect = ed.modelToView(beg);
                PopupMenu menu = new PopupMenu();
                if (suggestions.length > 0)
                {
                    ActionListener al = new ReplaceWordActionListener(ed, beg,
                                            end);
                    for (int i = 0; i < suggestions.length;
                            menu.add(new MenuItem(suggestions[i++]))
                                .addActionListener(al));
                }
                else
                {
                    MenuItem item = new MenuItem("(no spelling suggestions)");
                    item.setEnabled(false);
                    menu.add(item);
                }

                menu.addSeparator();

                MenuItem itemAdd = new MenuItem("Add to dictionary");
                itemAdd.addActionListener(new AddToDictionaryActionListener(word));
//                menu.add(itemIgnore);
                menu.add(itemAdd);
                ed.add(menu);
                SwingUtilities.invokeLater(new ShowNativeMenu(menu, ed, rect.x,
                        rect.y + rect.height, hi));
            }
            catch (BadLocationException ex)
            {
            }
            catch (Exception ex)
            {
            }
        }


        class ShowNativeMenu implements Runnable
        {
            private EditorPane editor;
            private int x, y;
            private PopupMenu menu;
            private Object highlight;

            ShowNativeMenu(PopupMenu m, EditorPane e, int x, int y, Object hi)
            {
                this.editor = e;
                this.x = x;
                this.y = y;
                this.menu = m;
                this.highlight = hi;
            }


            public void run()
            {
                menu.show(editor, x, y);
                editor.getHighlighter().removeHighlight(highlight);
            }
        }

    }


    class IgnoreAllActionListener extends SpellActionListener
    {
        private String word;

        IgnoreAllActionListener(String word)
        {
            this.word = word;
        }


        public void onAction(ActionEvent ev) throws SpellCheckerException
        {
            SpellCheckerTool.getChecker().ignore(word);
        }
    }


    class AddToDictionaryActionListener extends SpellActionListener
    {
        private String word;

        AddToDictionaryActionListener(String word)
        {
            this.word = word;
        }


        public void onAction(ActionEvent ev) throws SpellCheckerException
        {
            SpellCheckerTool.getChecker().addToDictionary(word);
        }
    }


    class ReplaceWordActionListener extends SpellActionListener
    {
        int p1, p2;
        EditorPane editor;

        ReplaceWordActionListener(EditorPane ed, int beg, int end)
        {
            p1 = beg;
            p2 = end;
            editor = ed;
        }


        public void onAction(ActionEvent ev) throws SpellCheckerException
        {
            JBUndoManager undo = ((EditorDocument)editor.getDocument()).getUndoManager();
            Caret caret = editor.getCaret();
            undo.startUndoGroup(caret.getDot(), caret.getMark());
            replaceWord(editor, p1, p2, ev.getActionCommand());
            undo.endUndoGroup(caret.getDot(), caret.getMark());
        }
    }


    /**
     * Action for checking entire file or when part of the text is selected - selection.
     * Dialog box with suggestions list is used in case of incorrectly spelled words.
     */
    public class CheckFileAction extends SpellAction
    {
        /**
         * Default constructor - passes action name and description to the base class.
         */
        public CheckFileAction()
        {
            super("check-file", "Check spelling of the current file.");
        }


        public void onAction(EditorPane editor) throws SpellCheckerException
        {
            Node n = Browser.getActiveBrowser().getActiveNode();
            if (!(n instanceof TextFileNode))
                return;

//            TextFileNode node = (TextFileNode)n;
//            node.

            try
            {
//                EditorPane editor = getEditorTarget(ev);
                Caret caret = editor.getCaret();
                int dot = caret.getDot();
                int mark = caret.getMark();
                if (dot == mark)
                {   //no selection - check everything
                    Document doc = editor.getDocument();
                    checkText(editor, doc.getStartPosition().getOffset(),
                              doc.getEndPosition().getOffset(), false);                }
                else
                {
                    //check only selected text
                    checkText(editor, Math.min(mark, dot), Math.max(mark, dot), true);
                }
            }
            catch (BadLocationException ex)
            {
                ex.printStackTrace();
            }
            catch (Exception ex)
            {
                ex.printStackTrace();
            }
        }


        private void checkText(EditorPane ed, int beg, int end, boolean selection)
            throws BadLocationException
        {
            new Thread(new TextChecker(ed, beg, end, selection)).start();
        }


        class TextChecker implements Runnable
        {
            private EditorPane ed;
            private int     beg;
            private int     end;
            private boolean selection;

            private SuggestionsDialog dlg;
            private DialogListener listener;
            private HashSet ignoreList = new HashSet();
            private HashMap replaceList = new HashMap();
            private boolean notAborted;
            private DefaultHighlighter.DefaultHighlightPainter painter =
                        new DefaultHighlighter.DefaultHighlightPainter(colorWrong);

            TextChecker(EditorPane ed, int beg, int end, boolean selection)
            {
                this.ed = ed;
                this.beg = beg;
                this.end = end;
                this.selection = selection;
            }


            public void run()
            {
                try
                {
                    JBUndoManager man = ((EditorDocument)ed.getDocument()).getUndoManager();
                    man.startUndoGroup(ed.getCaret().getDot(), ed.getCaret().getMark());

                    int p0 = TextUtilities.getPreviousWord(ed, beg);
                    int p1 = end;
                    notAborted = true;

                    while ((p0 < end) && (p0 != p1) && notAborted)
                    {
                        p1 = TextUtilities.getWordEnd(ed, p0, true);
                        p0 = TextUtilities.getWordStart(ed, p1, true);

                        checkWord(ed, p0, p1);

                        if (ed.getText(p1, 1).equals("_"))
                        {
                            p0 = p1 + 1;
                        }
                        else
                        {
                            p0 = TextUtilities.getNextWord(ed, p1);
                        }
                    }
                    if (notAborted && !Settings.noSpellCheckCompletedMessage.getBoolean())
                    {
                        boolean notAgain = MessageBox.showMessage((JFrame)ed.getTopLevelAncestor(),
                                        "Spell checking finished.", true);
                        Settings.noSpellCheckCompletedMessage.setBoolean(notAgain);
                    }

                    man.endUndoGroup(ed.getCaret().getDot(),
                                     ed.getCaret().getMark());
                }
                catch (BadLocationException ex)
                {
                }
                catch (SpellCheckerException ex)
                {
                    JOptionPane.showMessageDialog(Browser.getActiveBrowser(), ex.getMessage());
                    ex.tryToRecover();
                }
                finally
                {
                    if (dlg != null)
                    {
                        dlg.dispose();
                        dlg = null;
                    }
                }
            }


            private boolean isToBeChecked(EditorDocument d, int location)
                throws BadLocationException
            {
                EditorDocument$StyledLeafElement line =
                    (EditorDocument$StyledLeafElement)d
                        .getParagraphElement(location);
                Style style = line.getStyle(location);
                BasicStyleMap map = d.getStyleMap();

                if (style.getName().equals(map.getStyle(BasicStyleMap.COMMENT).getName()))
                {
                    return Settings.checkComments.getBoolean()
                            && !(Settings.ignoreJavaDocTags.getBoolean()
                                 && isJavaDocTag(ed, location));
                }

                if (style.getName().equals(map.getStyle(BasicStyleMap.STRING).getName()))
                {
                    return Settings.checkStrings.getBoolean();
                }

                return Settings.checkEverything.getBoolean()
                        || (Settings.checkEverythingInSelections.getBoolean() && selection);
            }


            private void checkWord(EditorPane ed, int beg, int end)
                throws BadLocationException, SpellCheckerException
            {
                EditorDocument d = (EditorDocument)ed.getDocument();
                if (!isToBeChecked(d, beg))
                {
                    return;
                }

                String word = ed.getText(beg, end - beg).trim();
                word = skipDigits(word);

//                System.err.println("checking \"" + word + "\"");

                AbstractChecker checker = SpellCheckerTool.getChecker();

                if (ignoreList.contains(word))
                {
                    return;
                }
                else if (replaceList.containsKey(word))
                {
                    replaceWord(ed, beg, end, (String)replaceList.get(word));
                }
                else if (word.length() > Settings.minWordLength.getInteger()
                         && !checker.isInDictionary(word))
                {
                    Rectangle rect = ed.modelToView(beg);
                    rect.width = ed.modelToView(end).x - rect.x;

                    //	ep.setTopLine( ep.getLineNumber( ep.getCaretPosition() ) );
                    ed.scrollRectToVisible(rect);
//                    ed.gotoLine(ed.getLineNumber(beg), false, EditorPane.ONETHIRD_IF_OFFSCREEN);

                    Object hi = ed.getHighlighter().addHighlight(beg, end, painter);
                    if (dlg == null)
                    {
                        startDialog((Frame)ed.getTopLevelAncestor());
                        dlg.setVisible(true);
                    }

                    Point p = rect.getLocation();
                    SwingUtilities.convertPointToScreen(p, ed);
                    ensureRectIsVisible(new Rectangle(p.x, p.y, rect.width, rect.height), dlg);
                    dlg.setSuggestions(word, checker.getSuggestions(word));
                    listener.setCurrentWord(beg, end, ed);

                    synchronized (listener)
                    {
                        try
                        {
                            listener.wait();
                            ed.getHighlighter().removeHighlight(hi);
                        }
                        catch (InterruptedException ex)
                        {
                        }
                        catch (Exception ex)
                        {
                            ex.printStackTrace();
                        }
                    }

                }
                else
                {
                    if (Settings.highlightAllCheckedWord.getBoolean())
                    {
//                        ed.gotoLine(ed.getLineNumber(beg), false, EditorPane.ONETHIRD_IF_OFFSCREEN);
                        temporarilyHighlightText(ed, beg, end, colorGood);
                    }
                }
            }


            private void ensureRectIsVisible(Rectangle rect, JDialog dlg)
            {
                Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
                Rectangle dlgRect = dlg.getBounds();
                if (dlgRect.intersects(rect))
                {
                    dlg.setBounds(dlgRect.x,
                                  (screen.height - rect.y - rect.height) > dlgRect.height
                                      ? rect.y + rect.height + 2
                                      : rect.y - dlgRect.height - 2,
                                  dlgRect.width,
                                  dlgRect.height);
                }
            }


            private void startDialog(Frame f)
            {
                listener = new DialogListener();
                dlg = new SuggestionsDialog(f, listener);

                Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
                Rectangle dlgRect = dlg.getBounds();
                int x = screen.width - dlgRect.width;
                int y = (screen.height - dlgRect.height) /2;
                dlg.setBounds(x, y, dlgRect.width, dlgRect.height);
            }


            class DialogListener implements SuggestionsDialog.DialogListener
            {
                private int beg;
                private int end;
                private EditorPane editor;

                public void setCurrentWord(int beg, int end, EditorPane editor)
                {
                    this.beg = beg;
                    this.end = end;
                    this.editor = editor;
                }


                public void onReplace(String oldWord, String newWord)
                    throws SpellCheckerException
                {
                    replaceWord(editor, beg, end, newWord);
                    onIgnore(oldWord);
                }


                /**
                 * Replace word with the new one, put the word in replaceList so
                 * it can be replaced automatically next time.
                 *
                 * @param oldWord
                 * @param newWord
                 */
                public void onReplaceAll(String oldWord, String newWord)
                    throws SpellCheckerException
                {
                    replaceList.put(oldWord, newWord);
                    onReplace(oldWord, newWord);
                }


                public void onIgnore(String word)
                    throws SpellCheckerException
                {
                    synchronized (listener)
                    {
                        listener.notify();
                    }
                }


                public void onIgnoreAll(String word)
                    throws SpellCheckerException
                {
                    ignoreList.add(word);
                    onIgnore(word);
                }


                public void onAddToDictionary(String word)
                    throws SpellCheckerException
                {
                    SpellCheckerTool.getChecker().addToDictionary(word);
                    onIgnore(word);
                }


                public void onAbortChecking()
                    throws SpellCheckerException
                {
                    notAborted = false;

                    onIgnore("");
                }
            }
        }

    }


    class UnderlineHighlightPainter extends LayeredHighlighter.LayerPainter
    {
        private Color color;

        UnderlineHighlightPainter(Color color)
        {
            this.color = color;
        }


        public Shape paintLayer(Graphics g, int p0, int p1, Shape s,
                                JTextComponent editor, View v)
        {
            try
            {
                g.setColor(color);

                Shape s0 = v.modelToView(p0, s, Position.Bias.Backward);
                Shape s1 = v.modelToView(p1, s, Position.Bias.Forward);
                int x0 = s0.getBounds().x;
                int x1 = s1.getBounds().x + s1.getBounds().width;
                int y = s0.getBounds().y + s0.getBounds().height - 2;
                for (boolean down = false; x0 < x1; x0 += 2)
                {
                    g.drawLine(x0, y, x0 + 1, y + (down ? -1 : 1));
                    down = !down;
                }

                g.drawLine(x0, y, x1, y);

                return s;
            }
            catch (Exception ex)
            {
                ex.printStackTrace();
            }

            return null;
        }


        public void paint(Graphics g, int offs0, int offs1, Shape s,
                          JTextComponent c)
        {
        }
    }


    class FrameHighlightPainter extends LayeredHighlighter.LayerPainter
    {
        private Color color;
        private Stroke stroke;

        FrameHighlightPainter(Color color)
        {
            this.color = color;
            // float dash1[] = {};
            stroke = new BasicStroke(2.0f);    // , BasicStroke.CAP_BUTT,
            // BasicStroke.JOIN_ROUND, 1.0f, dash1, 0.0f);
        }


        public Shape paintLayer(Graphics g, int p0, int p1, Shape s,
                                JTextComponent editor, View v)
        {
            try
            {
                Graphics2D g2 = (Graphics2D)g;
                g.setColor(color);

                Shape s0 = v.modelToView(p0, s, Position.Bias.Backward);
                Shape s1 = v.modelToView(p1, s, Position.Bias.Forward);
                int x0 = s0.getBounds().x - 1;
                int w = s1.getBounds().x - x0 + 1;
                int y0 = s0.getBounds().y;
                int h = s0.getBounds().height - 1;
                Stroke oldStroke = g2.getStroke();
                g2.setStroke(stroke);
                g.drawRect(x0, y0, w, h);
                g2.setStroke(oldStroke);

                return new Rectangle(x0 - 4, y0 - 4, w + 4, h + 4);
            }
            catch (Exception ex)
            {
                ex.printStackTrace();
            }

            return null;
        }


        public void paint(Graphics g, int offs0, int offs1, Shape s,
                          JTextComponent c)
        {
        }
    }

}


